import functools
import warnings
from collections import OrderedDict, defaultdict, namedtuple

import torch.autograd.profiler as torch_profiler

from .display import traces_to_display

Trace = namedtuple("Trace", ["path", "leaf", "module"])


def walk_modules(module, name="", path=()):
    """Generator. Walks through a PyTorch Module and outputs Trace tuples"""
    if not name:
        name = module.__class__.__name__
    named_children = list(module.named_children())
    path = path + (name,)
    yield Trace(path, len(named_children) == 0, module)
    # recursively walk into all submodules
    for name, child_module in named_children:
        yield from walk_modules(child_module, name=name, path=path)


class Profile(object):
    """Layer by layer profiling of PyTorch models, using the PyTorch autograd profiler."""

    def __init__(
        self, model, enabled=True, use_cuda=False, profile_memory=False, paths=None
    ):
        self._model = model
        self.enabled = enabled
        self.use_cuda = use_cuda
        self.profile_memory = profile_memory
        self.paths = paths

        self.entered = False
        self.exited = False
        self.traces = ()
        self._ids = set()
        self.trace_profile_events = defaultdict(list)

    def __enter__(self):
        if not self.enabled:
            return self
        if self.entered:
            raise RuntimeError("torchprof profiler is not reentrant")
        self.entered = True
        self._forwards = {}  # store the original forward functions
        self.traces = tuple(map(self._hook_trace, walk_modules(self._model)))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.enabled:
            return
        tuple(map(self._remove_hook_trace, self.traces))
        del self._forwards  # remove unnecessary forwards
        self.exited = True

    def __str__(self):
        return self.display()

    def __call__(self, *args, **kwargs):
        return self._model(*args, **kwargs)

    def _hook_trace(self, trace):
        [path, leaf, module] = trace
        _id = id(module)
        if (self.paths is not None and path in self.paths) or (
            self.paths is None and leaf
        ):
            if _id in self._ids:
                #already wrapped
                return trace
            self._ids.add(_id)
            _forward = module.forward
            self._forwards[path] = _forward

            @functools.wraps(_forward)
            def wrap_forward(*args, **kwargs):
                try:
                    with torch_profiler.profile(
                        use_cuda=self.use_cuda, profile_memory=self.profile_memory
                    ) as prof:
                        res = _forward(*args, **kwargs)
                except TypeError:
                    if self.profile_memory:
                        warnings.warn(
                            "`profile_memory` is unsupported in torch < 1.6",
                            RuntimeWarning,
                        )
                        self.profile_memory = False
                    with torch_profiler.profile(use_cuda=self.use_cuda) as prof:
                        res = _forward(*args, **kwargs)

                event_list = prof.function_events
                # PyTorch up until version 1.7 exposes this method. From PyTorch 1.8 onwards, 
                # it is called via EventList._build_tree at the end of the context manager.
                if hasattr(event_list, "populate_cpu_children"):
                    event_list.populate_cpu_children()
                # each profile call should be contained in its own list
                self.trace_profile_events[path].append(event_list)
                return res

            module.forward = wrap_forward
        return trace

    def _remove_hook_trace(self, trace):
        [path, leaf, module] = trace
        _id = id(module)
        if _id in self._ids:
            self._ids.discard(_id)
        else:
            return
        if (self.paths is not None and path in self.paths) or (
            self.paths is None and leaf
        ):
            module.forward = self._forwards[path]

    def raw(self):
        if self.exited:
            return (self.traces, self.trace_profile_events)

    def display(self, show_events=False):
        if self.exited:
            return traces_to_display(
                self.traces,
                self.trace_profile_events,
                show_events=show_events,
                paths=self.paths,
                use_cuda=self.use_cuda,
                profile_memory=self.profile_memory,
            )
        return "<unfinished torchprof.profile>"



import unittest
import torch
import torchprof
import torchvision


class TestSetPaths(unittest.TestCase):
    def test_cpu_profile_structure(self):
        model = torchvision.models.alexnet(pretrained=False)
        x = torch.rand([1, 3, 224, 224])

        paths = [("AlexNet", "features", "3"), ("AlexNet", "avgpool")]

        with torchprof.Profile(model, paths=paths) as prof:
            model(x)

        # print(prof)
        traces, event_dict = prof.raw()
        self.assertEqual(len(event_dict.keys()), 2)
        self.assertEqual(list(event_dict.keys()), paths)



import unittest
import torch
import torchprof
import torchvision


class TestProfile(unittest.TestCase):
    # ensure traces match alexnet
    alexnet_traces = [
        (("AlexNet",), False),
        (("AlexNet", "features"), False),
        (("AlexNet", "features", "0"), True),
        (("AlexNet", "features", "1"), True),
        (("AlexNet", "features", "2"), True),
        (("AlexNet", "features", "3"), True),
        (("AlexNet", "features", "4"), True),
        (("AlexNet", "features", "5"), True),
        (("AlexNet", "features", "6"), True),
        (("AlexNet", "features", "7"), True),
        (("AlexNet", "features", "8"), True),
        (("AlexNet", "features", "9"), True),
        (("AlexNet", "features", "10"), True),
        (("AlexNet", "features", "11"), True),
        (("AlexNet", "features", "12"), True),
        (("AlexNet", "avgpool"), True),
        (("AlexNet", "classifier"), False),
        (("AlexNet", "classifier", "0"), True),
        (("AlexNet", "classifier", "1"), True),
        (("AlexNet", "classifier", "2"), True),
        (("AlexNet", "classifier", "3"), True),
        (("AlexNet", "classifier", "4"), True),
        (("AlexNet", "classifier", "5"), True),
        (("AlexNet", "classifier", "6"), True),
    ]

    alexnet_cpu_ops = [
        None,  # 0
        None,  # 1
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::contiguous",
            "aten::mkldnn_convolution",
            "aten::as_strided_",
        ),  # 2
        ("aten::relu_", "aten::threshold_"),  # 3
        (
            "aten::max_pool2d",
            "aten::max_pool2d_with_indices",
            "aten::contiguous",
        ),  # 4
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::contiguous",
            "aten::mkldnn_convolution",
            "aten::as_strided_",
        ),  # 5
        ("aten::relu_", "aten::threshold_"),  # 6
        (
            "aten::max_pool2d",
            "aten::max_pool2d_with_indices",
            "aten::contiguous",
        ),  # 7
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::contiguous",
            "aten::mkldnn_convolution",
            "aten::as_strided_",
        ),  # 8
        ("aten::relu_", "aten::threshold_"),  # 9
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::contiguous",
            "aten::mkldnn_convolution",
            "aten::as_strided_",
        ),  # 10
        ("aten::relu_", "aten::threshold_"),  # 11
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::contiguous",
            "aten::mkldnn_convolution",
            "aten::as_strided_",
        ),  # 12
        ("aten::relu_", "aten::threshold_"),  # 13
        (
            "aten::max_pool2d",
            "aten::max_pool2d_with_indices",
            "aten::contiguous",
        ),  # 14
        (
            "aten::adaptive_avg_pool2d",
            "aten::_adaptive_avg_pool2d",
        ),  # 15
        None,  # 16
        (
            "aten::dropout",
            "aten::empty_like",
            "aten::bernoulli_",
            "aten::div_",
            "aten::to",
            "aten::empty_strided",
            "aten::mul",
        ),  # 17
        (
            "aten::t",
            "aten::transpose",
            "aten::as_strided",
            "aten::addmm",
            "aten::expand",
            "aten::as_strided",
        ),  # 18
        ("aten::relu_", "aten::threshold_"),  # 19
        (
            "aten::dropout",
            "aten::empty_like",
            "aten::bernoulli_",
            "aten::div_",
            "aten::to",
            "aten::empty_strided",
            "aten::mul",
        ),  # 20
        (
            "aten::t",
            "aten::transpose",
            "aten::as_strided",
            "aten::addmm",
            "aten::expand",
            "aten::as_strided",
        ),  # 21
        ("aten::relu_", "aten::threshold_"),  # 22
        (
            "aten::t",
            "aten::transpose",
            "aten::as_strided",
            "aten::addmm",
            "aten::expand",
            "aten::as_strided",
        ),  # 23
    ]

    alexnet_gpu_ops = (
        None,  # 0
        None,  # 1
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::cudnn_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::reshape",
            "aten::view",
            "aten::add_",
        ),  # 2
        ("aten::relu_", "aten::threshold_"),  # 3
        (
            "aten::max_pool2d",
            "aten::max_pool2d_with_indices",
            "aten::contiguous",
        ),  # 4
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::cudnn_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::reshape",
            "aten::view",
            "aten::add_",
        ),  # 5
        ("aten::relu_", "aten::threshold_"),  # 6
        (
            "aten::max_pool2d",
            "aten::max_pool2d_with_indices",
            "aten::contiguous",
        ),  # 7
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::cudnn_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::reshape",
            "aten::view",
            "aten::add_",
        ),  # 8
        ("aten::relu_", "aten::threshold_"),  # 9
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::cudnn_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::reshape",
            "aten::view",
            "aten::add_",
        ),  # 10
        ("aten::relu_", "aten::threshold_"),  # 11
        (
            "aten::conv2d",
            "aten::convolution",
            "aten::_convolution",
            "aten::contiguous",
            "aten::cudnn_convolution",
            "aten::contiguous",
            "aten::contiguous",
            "aten::reshape",
            "aten::view",
            "aten::add_",
        ),  # 12
        ("aten::relu_", "aten::threshold_"),  # 13
        (
            "aten::max_pool2d",
            "aten::max_pool2d_with_indices",
            "aten::contiguous",
        ),  # 14
        (
            "aten::adaptive_avg_pool2d",
            "aten::_adaptive_avg_pool2d",
            "aten::contiguous",
        ),  # 15
        None,  # 16
        (
            "aten::dropout",
            "aten::_fused_dropout",
            "aten::empty_like",
        ),  # 17
        (
            "aten::t",
            "aten::transpose",
            "aten::as_strided",
            "aten::addmm",
            "aten::expand",
            "aten::as_strided",
        ),  # 18
        ("aten::relu_", "aten::threshold_"),  # 19
        (
            "aten::dropout",
            "aten::_fused_dropout",
            "aten::empty_like",
        ),  # 20
        (
            "aten::t",
            "aten::transpose",
            "aten::as_strided",
            "aten::addmm",
            "aten::expand",
            "aten::as_strided",
        ),  # 21
        ("aten::relu_", "aten::threshold_"),  # 22
        (
            "aten::t",
            "aten::transpose",
            "aten::as_strided",
            "aten::addmm",
            "aten::expand",
            "aten::as_strided",
        ),  # 23
    )

    def test_cpu_profile_structure(self):
        model = torchvision.models.alexnet(pretrained=False)
        x = torch.rand([1, 3, 224, 224])
        self._profile_structure(model, x, alexnet_ops=self.alexnet_cpu_ops)

    @unittest.skipIf(not torch.cuda.is_available(), "CUDA not available")
    def test_cuda_profile_structure(self):
        model = torchvision.models.alexnet(pretrained=False).cuda()
        x = torch.rand([1, 3, 224, 224]).cuda()
        self._profile_structure(
            model, x, use_cuda=True, alexnet_ops=self.alexnet_gpu_ops
        )

    def _profile_structure(self, model, x, use_cuda=False, alexnet_ops=[]):
        with torchprof.Profile(model, use_cuda=use_cuda) as prof:
            model(x)

        traces, event_lists_dict = prof.raw()

        for layer_idx, trace in enumerate(traces):
            (path, leaf, _) = trace
            self.assertEqual((path, leaf), self.alexnet_traces[layer_idx])
            event_lists = event_lists_dict[path]
            if leaf:
                # model(x) called once, each layer should have one event_list
                self.assertEqual(len(event_lists), 1)
                event_names = tuple(e.name for e in event_lists[0])
                # profiler returned order is not deterministic
                try:
                    self.assertTrue(
                        all(event_name in event_names for event_name in alexnet_ops[layer_idx]),
                        f"Layer {layer_idx} received {event_names}, old {alexnet_ops[layer_idx]}",
                    )
                except IndexError:
                    self.assertTrue(False, f"Layer {layer_idx} received {event_names}")
            else:
                # non leaf nodes should not have event_list values
                self.assertEqual(len(event_lists), 0)

        pretty = prof.display()
        pretty_full = prof.display(show_events=True)
        self.assertIsInstance(pretty, str)
        self.assertIsInstance(pretty_full, str)

        # pprint.pprint(pretty)



from collections import OrderedDict, namedtuple

import torch.autograd.profiler as torch_profiler

Measure = namedtuple(
    "Measure",
    [  # when attr value is None, profiler unsupported
        "self_cpu_total",
        "cpu_total",
        "self_cuda_total",
        "cuda_total",
        "self_cpu_memory",
        "cpu_memory",
        "self_cuda_memory",
        "cuda_memory",
        "occurrences",
    ],
)


def _flatten_tree(t, depth=0):
    flat = []
    for name, st in t.items():
        measures = st.pop(None, None)
        flat.append([depth, name, measures])
        flat.extend(_flatten_tree(st, depth=depth + 1))
    return flat


def _build_measure_tuple(events, occurrences):
    # Events may have missing attributes depending on the PyTorch version used.

    # memory profiling supported in torch >= 1.6
    self_cpu_memory = None
    has_self_cpu_memory = any(hasattr(e, "self_cpu_memory_usage") for e in events)
    if has_self_cpu_memory:
        self_cpu_memory = sum([getattr(e, "self_cpu_memory_usage", 0) for e in events])
    cpu_memory = None
    has_cpu_memory = any(hasattr(e, "cpu_memory_usage") for e in events)
    if has_cpu_memory:
        cpu_memory = sum([getattr(e, "cpu_memory_usage", 0) for e in events])
    self_cuda_memory = None
    has_self_cuda_memory = any(hasattr(e, "self_cuda_memory_usage") for e in events)
    if has_self_cuda_memory:
        self_cuda_memory = sum(
            [getattr(e, "self_cuda_memory_usage", 0) for e in events]
        )
    cuda_memory = None
    has_cuda_memory = any(hasattr(e, "cuda_memory_usage") for e in events)
    if has_cuda_memory:
        cuda_memory = sum([getattr(e, "cuda_memory_usage", 0) for e in events])

    # self CUDA time supported in torch >= 1.7
    self_cuda_total = None
    has_self_cuda_time = any(hasattr(e, "self_cuda_time_total") for e in events)
    if has_self_cuda_time:
        self_cuda_total = sum([getattr(e, "self_cuda_time_total", 0) for e in events])

    return Measure(
        self_cpu_total=sum([e.self_cpu_time_total for e in events]),
        cpu_total=sum([e.cpu_time_total for e in events]),
        self_cuda_total=self_cuda_total,
        cuda_total=sum([e.cuda_time_total for e in events]),
        self_cpu_memory=self_cpu_memory,
        cpu_memory=cpu_memory,
        self_cuda_memory=self_cuda_memory,
        cuda_memory=cuda_memory,
        occurrences=occurrences,
    )


def _format_measure_tuple(measure):
    format_memory = getattr(torch_profiler, "format_memory", lambda _: "N/A")

    self_cpu_total = (
        torch_profiler.format_time(measure.self_cpu_total) if measure else ""
    )
    cpu_total = torch_profiler.format_time(measure.cpu_total) if measure else ""
    self_cuda_total = (
        torch_profiler.format_time(measure.self_cuda_total)
        if measure and measure.self_cuda_total is not None
        else ""
    )
    cuda_total = torch_profiler.format_time(measure.cuda_total) if measure else ""
    self_cpu_memory = (
        format_memory(measure.self_cpu_memory)
        if measure and measure.self_cpu_memory is not None
        else ""
    )
    cpu_memory = (
        format_memory(measure.cpu_memory)
        if measure and measure.cpu_memory is not None
        else ""
    )
    self_cuda_memory = (
        format_memory(measure.self_cuda_memory)
        if measure and measure.self_cuda_memory is not None
        else ""
    )
    cuda_memory = (
        format_memory(measure.cuda_memory)
        if measure and measure.cuda_memory is not None
        else ""
    )
    occurrences = str(measure.occurrences) if measure else ""

    return Measure(
        self_cpu_total=self_cpu_total,
        cpu_total=cpu_total,
        self_cuda_total=self_cuda_total,
        cuda_total=cuda_total,
        self_cpu_memory=self_cpu_memory,
        cpu_memory=cpu_memory,
        self_cuda_memory=self_cuda_memory,
        cuda_memory=cuda_memory,
        occurrences=occurrences,
    )


def group_by(events, keyfn):
    event_groups = OrderedDict()
    for event in events:
        key = keyfn(event)
        key_events = event_groups.get(key, [])
        key_events.append(event)
        event_groups[key] = key_events
    return event_groups.items()


def traces_to_display(
    traces,
    trace_events,
    show_events=False,
    paths=None,
    use_cuda=False,
    profile_memory=False,
    # dt = ('|', '|-- ', '+-- ', ' ') # ascii
    dt=("\u2502", "\u251c\u2500\u2500 ", "\u2514\u2500\u2500 ", " "),  # ascii-ex
):
    """Construct human readable output of the profiler traces and events."""
    tree = OrderedDict()

    for trace in traces:
        [path, leaf, module] = trace
        current_tree = tree
        # unwrap all of the events, in case model is called multiple times
        events = [te for t_events in trace_events[path] for te in t_events]
        for depth, name in enumerate(path, 1):
            if name not in current_tree:
                current_tree[name] = OrderedDict()
            if depth == len(path) and (
                (paths is None and leaf) or (paths is not None and path in paths)
            ):
                # tree measurements have key None, avoiding name conflict
                if show_events:
                    for event_name, event_group in group_by(events, lambda e: e.name):
                        event_group = list(event_group)
                        current_tree[name][event_name] = {
                            None: _build_measure_tuple(event_group, len(event_group))
                        }
                else:
                    current_tree[name][None] = _build_measure_tuple(
                        events, len(trace_events[path])
                    )
            current_tree = current_tree[name]
    tree_lines = _flatten_tree(tree)

    format_lines = []
    has_self_cuda_total = False
    has_self_cpu_memory = False
    has_cpu_memory = False
    has_self_cuda_memory = False
    has_cuda_memory = False

    for idx, tree_line in enumerate(tree_lines):
        depth, name, measures = tree_line

        next_depths = [pl[0] for pl in tree_lines[idx + 1 :]]
        pre = ""
        if depth > 0:
            pre = dt[1] if depth in next_depths and next_depths[0] >= depth else dt[2]
            depth -= 1
        while depth > 0:
            pre = (dt[0] + pre) if depth in next_depths else (dt[3] + pre)
            depth -= 1
        format_lines.append([pre + name, *_format_measure_tuple(measures)])
        if measures:
            has_self_cuda_total = (
                has_self_cuda_total or measures.self_cuda_total is not None
            )
            has_self_cpu_memory = (
                has_self_cpu_memory or measures.self_cpu_memory is not None
            )
            has_cpu_memory = has_cpu_memory or measures.cpu_memory is not None
            has_self_cuda_memory = (
                has_self_cuda_memory or measures.self_cuda_memory is not None
            )
            has_cuda_memory = has_cuda_memory or measures.cuda_memory is not None

    # construct the table (this is pretty ugly and can probably be optimized)
    heading = (
        "Module",
        "Self CPU total",
        "CPU total",
        "Self CUDA total",
        "CUDA total",
        "Self CPU Mem",
        "CPU Mem",
        "Self CUDA Mem",
        "CUDA Mem",
        "Number of Calls",
    )
    max_lens = [max(map(len, col)) for col in zip(*([heading] + format_lines))]

    # not all columns should be displayed, specify kept indexes
    keep_indexes = [0, 1, 2, 9]
    if profile_memory:
        if has_self_cpu_memory:
            keep_indexes.append(5)
        if has_cpu_memory:
            keep_indexes.append(6)
    if use_cuda:
        if has_self_cuda_total:
            keep_indexes.append(3)
        keep_indexes.append(4)
        if profile_memory:
            if has_self_cuda_memory:
                keep_indexes.append(7)
            if has_cuda_memory:
                keep_indexes.append(8)
    keep_indexes = tuple(sorted(keep_indexes))

    display = (  # table heading
        " | ".join(
            [
                "{:<{}s}".format(heading[keep_index], max_lens[keep_index])
                for keep_index in keep_indexes
            ]
        )
        + "\n"
    )
    display += (  # separator
        "-|-".join(
            [
                "-" * max_len
                for val_idx, max_len in enumerate(max_lens)
                if val_idx in keep_indexes
            ]
        )
        + "\n"
    )
    for format_line in format_lines:  # body
        display += (
            " | ".join(
                [
                    "{:<{}s}".format(value, max_lens[val_idx])
                    for val_idx, value in enumerate(format_line)
                    if val_idx in keep_indexes
                ]
            )
            + "\n"
        )

    return display



import codecs
import os
import re
import setuptools

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


long_description = read("README.md")
version = find_version("torchprof", "__init__.py")

setuptools.setup(
    name="torchprof",
    version=version,
    author="Alexander Wong",
    author_email="alex@udia.ca",
    description="Measure neural network device specific metrics (latency, flops, etc.)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/awwong1/torchprof",
    packages=setuptools.find_packages(),
    license="MIT",
    install_requires=["torch>=1.1.0,<2"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
